@@ -1,6 +1,6 @@
 /* cfftot1.cc -- driver for translating CFF fonts to Type 1 fonts
  *
- * Copyright (c) 2002-2009 Eddie Kohler
+ * Copyright (c) 2002-2010 Eddie Kohler
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the Free
@@ -127,7 +127,7 @@ do_file(const char *infn, const char *outfn, PermString name, ErrorHandler *errh
 	StringAccum sa(150000);
 	while (!feof(f)) {
 	    int forward = fread(sa.reserve(32768), 1, 32768, f);
-	    sa.forward(forward);
+	    sa.adjust_length(forward);
 	}
 	if (f != stdin)
 	    fclose(f);
@@ -210,7 +210,7 @@ main(int argc, char *argv[])
 
 	  case VERSION_OPT:
 	    printf("cfftot1 (LCDF typetools) %s\n", VERSION);
-	    printf("Copyright (C) 2002-2009 Eddie Kohler\n\
+	    printf("Copyright (C) 2002-2010 Eddie Kohler\n\
 This is free software; see the source for copying conditions.\n\
 There is NO warranty, not even for merchantability or fitness for a\n\
 particular purpose.\n");
