@@ -4,8 +4,8 @@
  * Eddie Kohler
  *
  * Copyright (c) 1999-2000 Massachusetts Institute of Technology
- * Copyright (c) 2001-2008 Eddie Kohler
- * Copyright (c) 2008 Meraki, Inc.
+ * Copyright (c) 2001-2010 Eddie Kohler
+ * Copyright (c) 2008-2009 Meraki, Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
  * copy of this software and associated documentation files (the "Software"),
@@ -69,43 +69,61 @@
  * returns an out-of-memory string.
  */
 
-const char String::null_string_data = 0;
-const char String::oom_string_data = 0;
+const char String::null_data = '\0';
+const char String::oom_data = '\0';
 const char String::bool_data[] = "true\0false";
+const char String::int_data[] = "0\0001\0002\0003\0004\0005\0006\0007\0008\0009";
 
-String::memo_t String::null_memo = {
-    2, 0, 0, const_cast<char *>(&null_string_data)
-};
-String::memo_t String::permanent_memo = {
-    1, 0, 0, const_cast<char *>(&null_string_data)
-};
-String::memo_t String::oom_memo = {
-    2, 0, 0, const_cast<char *>(&oom_string_data)
-};
+#if HAVE_STRING_PROFILING > 1
+# define MEMO_INITIALIZER_TAIL , 0, 0
+#else
+# define MEMO_INITIALIZER_TAIL
+#endif
 
 const String::rep_t String::null_string_rep = {
-    &null_string_data, 0, &null_memo
+    &null_data, 0, 0
 };
 const String::rep_t String::oom_string_rep = {
-    &oom_string_data, 0, &oom_memo
+    &oom_data, 0, 0
 };
 
+#if HAVE_STRING_PROFILING
+uint64_t String::live_memo_count;
+uint64_t String::memo_sizes[55];
+uint64_t String::live_memo_sizes[55];
+uint64_t String::live_memo_bytes[55];
+# if HAVE_STRING_PROFILING > 1
+String::memo_t *String::live_memos[55];
+# endif
+#endif
+
 /** @cond never */
 String::memo_t *
-String::create_memo(char *data, int dirty, int capacity)
+String::create_memo(char *space, int dirty, int capacity)
 {
-    assert(capacity >= dirty);
-    memo_t *memo = new memo_t;
+    assert(capacity > 0 && capacity >= dirty);
+    memo_t *memo;
+    if (space)
+	memo = reinterpret_cast<memo_t *>(space);
+    else
+	memo = reinterpret_cast<memo_t *>(new char[MEMO_SPACE + capacity]);
     if (memo) {
-	if (data)
-	    memo->real_data = data;
-	else if (!(memo->real_data = new char[capacity])) {
-	    delete memo;
-	    return 0;
-	}
 	memo->capacity = capacity;
 	memo->dirty = dirty;
-	memo->refcount = (data ? 0 : 1);
+	memo->refcount = (space ? 0 : 1);
+#if HAVE_STRING_PROFILING
+	int bucket = profile_memo_size_bucket(dirty, capacity);
+	++memo_sizes[bucket];
+	++live_memo_sizes[bucket];
+	live_memo_bytes[bucket] += capacity;
+	++live_memo_count;
+# if HAVE_STRING_PROFILING > 1
+	memo->pprev = &live_memos[bucket];
+	if ((memo->next = *memo->pprev))
+	    memo->next->pprev = &memo->next;
+	*memo->pprev = memo;
+# endif
+#endif
     }
     return memo;
 }
@@ -113,47 +131,121 @@ String::create_memo(char *data, int dirty, int capacity)
 void
 String::delete_memo(memo_t *memo)
 {
-    if (memo->capacity) {
-	assert(memo->capacity >= memo->dirty);
-	delete[] memo->real_data;
-    }
-    delete memo;
+    assert(memo->capacity > 0);
+    assert(memo->capacity >= memo->dirty);
+#if HAVE_STRING_PROFILING
+    int bucket = profile_memo_size_bucket(memo->dirty, memo->capacity);
+    --live_memo_sizes[bucket];
+    live_memo_bytes[bucket] -= memo->capacity;
+    --live_memo_count;
+# if HAVE_STRING_PROFILING > 1
+    if ((*memo->pprev = memo->next))
+	memo->next->pprev = memo->pprev;
+# endif
+#endif
+    delete[] reinterpret_cast<char *>(memo);
 }
+
+
+
+#if HAVE_STRING_PROFILING
+void
+String::one_profile_report(StringAccum &sa, int i, int examples)
+{
+    if (i <= 16)
+	sa << "memo_dirty_" << i;
+    else if (i < 25) {
+	uint32_t s = (i - 17) * 2 + 17;
+	sa << "memo_cap_" << s << '_' << (s + 1);
+    } else if (i < 29) {
+	uint32_t s = (i - 25) * 8 + 33;
+	sa << "memo_cap_" << s << '_' << (s + 7);
+    } else {
+	uint32_t s1 = (1U << (i - 23)) + 1;
+	uint32_t s2 = (s1 - 1) << 1;
+	sa << "memo_cap_" << s1 << '_' << s2;
+    }
+    sa << '\t' << live_memo_sizes[i] << '\t' << memo_sizes[i] << '\t' << live_memo_bytes[i] << '\n';
+    if (examples) {
+# if HAVE_STRING_PROFILING > 1
+	for (memo_t *m = live_memos[i]; m; m = m->next) {
+	    sa << "    [" << m->dirty << "] ";
+	    uint32_t dirty = m->dirty;
+	    if (dirty > 0 && m->real_data[dirty - 1] == '\0')
+		--dirty;
+	    sa.append(m->real_data, dirty > 128 ? 128 : dirty);
+	    sa << '\n';
+	}
+# endif
+    }
+}
+
+void
+String::profile_report(StringAccum &sa, int examples)
+{
+    uint64_t all_live_sizes = 0, all_sizes = 0, all_live_bytes = 0;
+    for (int i = 0; i < 55; ++i) {
+	if (memo_sizes[i])
+	    one_profile_report(sa, i, examples);
+	all_live_sizes += live_memo_sizes[i];
+	all_sizes += memo_sizes[i];
+	all_live_bytes += live_memo_bytes[i];
+    }
+    sa << "memo_total\t" << all_live_sizes << '\t' << all_sizes << '\t' << all_live_bytes << '\n';
+}
+#endif
+
 /** @endcond never */
 
 
-String::String(int i)
+String::String(int x)
 {
-    char buf[128];
-    sprintf(buf, "%d", i);
-    assign(buf, -1, false);
+    if (x >= 0 && x < 10)
+	assign_memo(int_data + 2 * x, 1, 0);
+    else {
+	char buf[128];
+	sprintf(buf, "%d", x);
+	assign(buf, -1, false);
+    }
 }
 
-String::String(unsigned u)
+String::String(unsigned x)
 {
-    char buf[128];
-    sprintf(buf, "%u", u);
-    assign(buf, -1, false);
+    if (x < 10)
+	assign_memo(int_data + 2 * x, 1, 0);
+    else {
+	char buf[128];
+	sprintf(buf, "%u", x);
+	assign(buf, -1, false);
+    }
 }
 
-String::String(long i)
+String::String(long x)
 {
-    char buf[128];
-    sprintf(buf, "%ld", i);
-    assign(buf, -1, false);
+    if (x >= 0 && x < 10)
+	assign_memo(int_data + 2 * x, 1, 0);
+    else {
+	char buf[128];
+	sprintf(buf, "%ld", x);
+	assign(buf, -1, false);
+    }
 }
 
-String::String(unsigned long u)
+String::String(unsigned long x)
 {
-    char buf[128];
-    sprintf(buf, "%lu", u);
-    assign(buf, -1, false);
+    if (x < 10)
+	assign_memo(int_data + 2 * x, 1, 0);
+    else {
+	char buf[128];
+	sprintf(buf, "%lu", x);
+	assign(buf, -1, false);
+    }
 }
 
-String::String(double d)
+String::String(double x)
 {
     char buf[128];
-    int len = sprintf(buf, "%.12g", d);
+    int len = sprintf(buf, "%.12g", x);
     assign(buf, len, false);
 }
 
@@ -161,10 +253,8 @@ String
 String::make_claim(char *str, int len, int capacity)
 {
     assert(str && len > 0 && capacity >= len);
-    if (memo_t *new_memo = create_memo(str, len, capacity))
-	return String(str, len, new_memo);
-    else
-	return String(&oom_string_data, 0, &oom_memo);
+    memo_t *new_memo = create_memo(str - MEMO_SPACE, len, capacity);
+    return String(str, len, new_memo);
 }
 
 String
@@ -172,10 +262,7 @@ String::make_stable(const char *s, int len)
 {
     if (len < 0)
 	len = (s ? strlen(s) : 0);
-    if (len == 0)
-	return String();
-    else
-	return String(s, len, &permanent_memo);
+    return String(s, len, 0);
 }
 
 String
@@ -199,10 +286,9 @@ String::assign_out_of_memory()
 {
     if (_r.memo)
 	deref();
-    _r.memo = &oom_memo;
-    _r.data = _r.memo->real_data;
+    _r.memo = 0;
+    _r.data = &oom_data;
     _r.length = 0;
-    ++_r.memo->refcount;
 }
 
 void
@@ -216,7 +302,8 @@ String::assign(const char *str, int len, bool need_deref)
 
     // need to start with dereference
     if (need_deref) {
-	if (str >= _r.memo->real_data
+	if (_r.memo
+	    && str >= _r.memo->real_data
 	    && str + len <= _r.memo->real_data + _r.memo->capacity) {
 	    // Be careful about "String s = ...; s = s.c_str();"
 	    _r.data = str;
@@ -227,21 +314,21 @@ String::assign(const char *str, int len, bool need_deref)
     }
 
     if (len == 0) {
-	_r.memo = (str == &oom_string_data ? &oom_memo : &null_memo);
-	++_r.memo->refcount;
+	_r.memo = 0;
+	_r.data = (str == &oom_data ? str : &null_data);
 
     } else {
-	// Make 'capacity' a multiple of 16 characters and bigger than 'len'.
-	int capacity = (len + 16) & ~15;
-	_r.memo = create_memo(0, len, capacity);
+	// Make the memo a multiple of 16 characters and bigger than 'len'.
+	int memo_capacity = (len + 15 + MEMO_SPACE) & ~15;
+	_r.memo = create_memo(0, len, memo_capacity - MEMO_SPACE);
 	if (!_r.memo) {
 	    assign_out_of_memory();
 	    return;
 	}
 	memcpy(_r.memo->real_data, str, len);
+	_r.data = _r.memo->real_data;
     }
 
-    _r.data = _r.memo->real_data;
     _r.length = len;
 }
 
@@ -249,32 +336,40 @@ char *
 String::append_garbage(int len)
 {
     // Appending anything to "out of memory" leaves it as "out of memory"
-    if (len <= 0 || _r.memo == &oom_memo)
+    if (len <= 0 || _r.data == &oom_data)
 	return 0;
 
     // If we can, append into unused space. First, we check that there's
     // enough unused space for 'len' characters to fit; then, we check
     // that the unused space immediately follows the data in '*this'.
-    uint32_t dirty = _r.memo->dirty;
-    if (_r.memo->capacity > dirty + len) {
+    uint32_t dirty;
+    if (_r.memo
+	&& ((dirty = _r.memo->dirty), _r.memo->capacity > dirty + len)) {
 	char *real_dirty = _r.memo->real_data + dirty;
 	if (real_dirty == _r.data + _r.length) {
-	    _r.length += len;
 	    _r.memo->dirty = dirty + len;
+	    _r.length += len;
 	    assert(_r.memo->dirty < _r.memo->capacity);
+#if HAVE_STRING_PROFILING
+	    profile_update_memo_dirty(_r.memo, dirty, dirty + len, _r.memo->capacity);
+#endif
 	    return real_dirty;
 	}
     }
 
-    // Now we have to make new space. Make sure the new capacity is a
-    // multiple of 16 characters and that it is at least 16. But for large
-    // strings, allocate a power of 2, since power-of-2 sizes minimize waste
-    // in frequently-used allocators, like Linux kmalloc.
-    int new_capacity = (_r.length + len < 1024 ? (_r.length + 16) & ~15 : 1024);
-    while (new_capacity < _r.length + len)
-	new_capacity *= 2;
+    // Now we have to make new space. Make sure the memo is a multiple of 16
+    // bytes and that it is at least 16. But for large strings, allocate a
+    // power of 2, since power-of-2 sizes minimize waste in frequently-used
+    // allocators, like Linux kmalloc.
+    int want_memo_len = _r.length + len + MEMO_SPACE;
+    int memo_capacity;
+    if (want_memo_len <= 1024)
+	memo_capacity = (want_memo_len + 15) & ~15;
+    else
+	for (memo_capacity = 2048; memo_capacity < want_memo_len; )
+	    memo_capacity *= 2;
 
-    memo_t *new_memo = create_memo(0, _r.length + len, new_capacity);
+    memo_t *new_memo = create_memo(0, _r.length + len, memo_capacity - MEMO_SPACE);
     if (!new_memo) {
 	assign_out_of_memory();
 	return 0;
@@ -300,13 +395,14 @@ String::append(const char *s, int len)
     } else if (len < 0)
 	len = strlen(s);
 
-    if (s == &oom_string_data)
+    if (s == &oom_data)
 	// Appending "out of memory" to a regular string makes it "out of
 	// memory"
 	assign_out_of_memory();
     else if (len == 0)
 	/* do nothing */;
-    else if (!(s >= _r.memo->real_data
+    else if (!(_r.memo
+	       && s >= _r.memo->real_data
 	       && s + len <= _r.memo->real_data + _r.memo->capacity)) {
 	if (char *space = append_garbage(len))
 	    memcpy(space, s, len);
@@ -330,12 +426,14 @@ String::mutable_data()
 {
     // If _memo has a capacity (it's not one of the special strings) and it's
     // uniquely referenced, return _data right away.
-    if (_r.memo->capacity && _r.memo->refcount == 1)
+    if (_r.memo && _r.memo->refcount == 1)
 	return const_cast<char *>(_r.data);
 
     // Otherwise, make a copy of it. Rely on: deref() doesn't change _data or
     // _length; and if _capacity == 0, then deref() doesn't free _real_data.
-    assert(!_r.memo->capacity || _r.memo->refcount > 1);
+    assert(!_r.memo || _r.memo->refcount > 1);
+    // But in multithreaded situations we must hold a local copy of memo!
+    String do_not_delete_underlying_memo(*this);
     deref();
     assign(_r.data, _r.length, false);
     return const_cast<char *>(_r.data);
@@ -349,53 +447,6 @@ String::mutable_c_str()
     return const_cast<char *>(_r.data);
 }
 
-const char *
-String::c_str() const
-{
-    // If _memo has no capacity, then this is one of the special strings (null
-    // or PermString). We are guaranteed, in these strings, that
-    // _data[_length] exists. We can return _data immediately if we have a
-    // '\0' in the right place.
-    if (!_r.memo->capacity && _r.data[_r.length] == '\0')
-	return _r.data;
-
-    // Otherwise, this invariant must hold (there's more real data in _memo
-    // than in our substring).
-    assert(!_r.memo->capacity
-	   || _r.memo->real_data + _r.memo->dirty >= _r.data + _r.length);
-
-    // Has the character after our substring been set?
-    uint32_t dirty = _r.memo->dirty;
-    if (_r.memo->real_data + dirty == _r.data + _r.length) {
-	if (_r.memo->capacity > dirty) {
-	    _r.memo->dirty = dirty + 1;
-	    // Character after our substring has not been set. Change it to
-	    // '\0'. This case will never occur on special strings.
-	    char *real_data = const_cast<char *>(_r.data);
-	    real_data[_r.length] = '\0';
-	    return _r.data;
-	}
-
-    } else {
-	// Character after our substring has been set. OK to return _data if
-	// it is already '\0'.
-	if (_r.data[_r.length] == '\0')
-	    return _r.data;
-    }
-
-    // If we get here, we must make a copy of our portion of the string.
-    {
-	String s(_r.data, _r.length);
-	deref();
-	assign(s);
-    }
-
-    char *real_data = const_cast<char *>(_r.data);
-    real_data[_r.length] = '\0';
-    ++_r.memo->dirty;		// include '\0' in used portion of _memo
-    return _r.data;
-}
-
 String
 String::substring(int pos, int len) const
 {
@@ -531,7 +582,7 @@ String::printable() const
     return *this;
 }
 
-uint32_t
+hashcode_t
 String::hashcode(const char *begin, const char *end)
 {
     if (end <= begin)
@@ -612,7 +663,7 @@ String::hashcode(const char *begin, const char *end)
 #if 0
 // 11.Apr.2008 -- This old hash function was swapped out in favor of
 // SuperFastHash, above.
-size_t
+hashcode_t
 String::hashcode() const
 {
     int l = length();
@@ -643,7 +694,7 @@ String::equals(const char *s, int len) const
     else if (_r.data == s)
 	return true;
     else if (len == 0)
-	return (s != &oom_string_data && _r.memo != &oom_memo);
+	return (s != &oom_data && _r.data != &oom_data);
     else
 	return memcmp(_r.data, s, len) == 0;
 }
@@ -659,7 +710,7 @@ String::starts_with(const char *s, int len) const
     else if (_r.data == s)
 	return true;
     else if (len == 0)
-	return (s != &oom_string_data && _r.memo != &oom_memo);
+	return (s != &oom_data && _r.data != &oom_data);
     else
 	return memcmp(_r.data, s, len) == 0;
 }
@@ -671,9 +722,9 @@ String::compare(const char *s, int len) const
 	len = strlen(s);
     if (_r.data == s)
 	return _r.length - len;
-    else if (_r.memo == &oom_memo)
+    else if (_r.data == &oom_data)
 	return 1;
-    else if (s == &oom_string_data)
+    else if (s == &oom_data)
 	return -1;
     else if (_r.length == len)
 	return memcmp(_r.data, s, len);
